//********************************************************************
//*** ResourcePool/IgnoreExceptionCommand.hh
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: IgnoreExceptionCommand.hh,v 1.2 2008/01/27 16:42:27 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#ifndef RESOURCEPOOL_IGNOREEXCEPTIONCOMMAND_HH
#define RESOURCEPOOL_IGNOREEXCEPTIONCOMMAND_HH

#include "mm/gc_ptr.hh"
#include "util/Clone.hh"

namespace fatalmind {

/**
 * This class takes another Command as argument and eats 
 * the exception if one is thrown.
 */

// TODO
//  * addtional Template Parameter to catch only specified exceptions
//  * make a getter for this exception
template<class CommandType>
class IgnoreExceptionCommand: public CommandType
{
	protected:
	private:
		gc_ptr<CommandType> m;

		IgnoreExceptionCommand(const IgnoreExceptionCommand& rho) 
		: CommandType(rho)
		, m(rho.m)
		{
		}
	public:
		IgnoreExceptionCommand(const CommandType& a_m) 
		: CommandType(a_m)
		, m(a_m.template clone<CommandType>())
		{
		};
		virtual ~IgnoreExceptionCommand() {
		}

		virtual void execute(typename CommandType::pool_type::resource_t& r) {
			try {
				m->execute(r);
			} catch (...) {
			}
		}

		const CommandType& getCommandRef() const {
			return *m.get();
		}

	protected:
		virtual Clone* DoClone() const {
			return new IgnoreExceptionCommand(*this);
		}

}; // class IgnoreExceptionCommand

} // namespace fatalmind
#endif
