//********************************************************************
//*** ResourcePool/oracle/SQLExecute.cc
//*** Copyright (c) 2003-2008 by Markus Winand <mws@fatalmind.com>
//*** $Id: SQLExecute.cc,v 1.28 2008/01/27 16:42:29 mws Exp $
//********************************************************************
/*
This file is part of ResourcePool.

ResourcePool is free software; you can redistribute it
and/or modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2 of 
the License, or (at your option) any later version.

ResourcePool is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ResourcePool; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307  USA */

#include "oci.h"
#include <algorithm>

#ifndef RESOURCEPOOL_ORACLE_SQLEXECUTE_HH
#include "SQLExecute.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCE_HH
#include "Resource.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_OCIHANDLE_HH
#include "OCIHandle.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLETYPE_HH
#include "oracleType.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_ORACLEEXCEPTIONFACTORY_HH
#include "OracleExceptionFactory.hh"
#endif

#ifndef RESOURCEPOOL_ORACLE_COMMANDINTERFACE_CC
#include "CommandInterface.cc"
#endif

#ifndef RESOURCEPOOL_ORACLE_RESOURCEPOOL_HH
#include "ResourcePool.hh"
#endif

namespace fatalmind {
namespace oracle {

template<class TM, template<class> class SLMDC>
class SQLExecute<TM, SLMDC>::SQLExecuteBatchArguementsImpl {
	public:
		std::vector<std::vector<ub2> > sizes;
		std::vector<std::vector<sb2> > indicators;
};

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const std::string& SQL)
: CommandCommon<TM, SLMDC>(SQL)
, batchParams()
, batchParamData(new SQLExecuteBatchArguementsImpl())
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const ResourcePool<ResourceType<TM> >& p, const std::string& SQL)
: CommandCommon<TM, SLMDC>(SQL)
, batchParams()
, batchParamData(new SQLExecuteBatchArguementsImpl())
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::SQLExecute(const this_t& rho) 
: CommandCommon<TM, SLMDC>(rho)
, batchParams(rho.batchParams)
, batchParamData(new SQLExecuteBatchArguementsImpl(*rho.batchParamData))
{
}

template<class TM, template<class> class SLMDC>
SQLExecute<TM, SLMDC>::~SQLExecute() 
{
}


template<class TM, template<class> class SLMDC>
Clone* SQLExecute<TM, SLMDC>::DoClone() const 
{
	return new SQLExecute<TM, SLMDC>(*this);
};


template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::_execute(oracleResource<TM>& rhand, StatementCache& sthcache) 
{
	oracleHandlesWithCache<TM>& hand = rhand.get_plain_resource();
	sword ret;
	prepare(hand, sthcache, oracleStatement(_SQL));

	if (batchSize() == 1) {
		real_bind(hand);
	} else {
		real_bind_batch(hand);
	}


	/* execute statement */
	if ((ret = OCIStmtExecute(hand.getSvcCtx()
		, _sth->get()
		, hand.getError()
		, (ub4) batchSize()
		, (ub4) 0
		, (CONST OCISnapshot *) 0
		, (OCISnapshot *) 0
		, (ub4) ((hand.getTxDepth() == 0) ? OCI_COMMIT_ON_SUCCESS : OCI_DEFAULT) | (batchSize() > 1 ? OCI_BATCH_ERRORS :  OCI_DEFAULT)) )
	) {
		if (batchSize() == 1) {
			OracleExceptionFactory::throwit("OCIStmtExecute", hand, ret, _SQL);
		} else {
			int row_offset = 42, num_errs = 42;
			OCIErrorHandle errhp2(hand.getEnv());

			OCIAttrGet (_sth->get(), OCI_HTYPE_STMT, &num_errs, 0,  OCI_ATTR_NUM_DML_ERRORS, errhp2.get());
			if (num_errs) {
				OCIError *errhndl[1];
				// TODO, use my own classes for that
				OCIHandleAlloc( hand.getEnv(), (dvoid **)&errhndl[0], (ub4) OCI_HTYPE_ERROR, 0, (dvoid **) 0);
				OCIParamGet(hand.getError(), OCI_HTYPE_ERROR, errhp2.get(), (void**)&errhndl[0], 0);

				OCIAttrGet(errhndl[0], OCI_HTYPE_ERROR, &row_offset, 0, OCI_ATTR_DML_ROW_OFFSET, errhp2.get());
				std::ostringstream oss;
				oss << "OCIStmtExecute(at batch position " << row_offset << " with " << num_errs << " errors)";
				OracleExceptionFactory::throwit(oss.str(), errhndl[0], ret,  OCI_HTYPE_ERROR, _SQL);
			}
		}
	} 
/*
else if (batchSize() > 1) {
			int row_offset = 42, num_errs = 42;
			OCIErrorHandle errhp2(hand.getEnv());
			OCIError *errhndl[1];
			OCIHandleAlloc( hand.getEnv(), (dvoid **)&errhndl[0], (ub4) OCI_HTYPE_ERROR, 0, (dvoid **) 0);

			OCIAttrGet (_sth->get(), OCI_HTYPE_STMT, &num_errs, 0,  OCI_ATTR_NUM_DML_ERRORS, errhp2.get());
			if (num_errs) {
				OCIParamGet(hand.getError(), OCI_HTYPE_ERROR, errhp2.get(), (void**)&errhndl[0], 0);

				OCIAttrGet(errhndl[0], OCI_HTYPE_ERROR, &row_offset, 0, OCI_ATTR_DML_ROW_OFFSET, errhp2.get());
				std::ostringstream oss;
				oss << "OCIStmtExecute(at batch position " << row_offset << " wiht " << num_errs << "errors)";
				OracleExceptionFactory::throwit(oss.str(), errhandl[1], ret, _SQL);
			}
	}
*/
}
template<class TM, template<class> class SLMDC>
void SQLExecute<TM, SLMDC>::real_bind_batch(oracleHandles& hand) {
	sword ret;
	typename super::_argst::const_iterator end = super::_args.end();
        int pos = 0;
	batchBuffers.clear();
        for (typename super::_argst::const_iterator i = super::_args.begin(); i != end; ++i) {
		// obtain maximum size
	
		int mxsize = *max_element(batchParamData->sizes[pos].begin(), batchParamData->sizes[pos].end());
		gc_ptr<RawBuffer> buf(new RawBuffer(mxsize*batchParams.size()));
		batchBuffers.push_back(buf); // todo, release
		
		typename batchParams_t::const_iterator tend = batchParams.end();	
		char *buffer = (char*)buf->get();
		for (typename batchParams_t::const_iterator t = batchParams.begin(); t != tend; ++t) {
			memcpy(buffer, (*t)[pos]->ptr(), (*t)[pos]->size());
			buffer += mxsize;
		}

                OCIBind  *bnd1p = NULL;
		if ((ret = OCIBindByPos(_sth->get()
			, &bnd1p
			, hand.getError()
			, pos+1
			, buf->get()
			, mxsize
			, (*i)->type()
			, (sb2 *)(&(batchParamData->indicators[pos][0]))
			, (ub2 *)(&(batchParamData->sizes[pos][0])) // alenp
			, (ub2 *)0
			, 0 // maxarr_len (only for PL/SQL)
			, (ub4 *)0
			, OCI_DEFAULT)
		) != OCI_SUCCESS) {
			OracleExceptionFactory::throwit("OCIBindByPos", hand, ret, _SQL);
		}


		// array it ;)
		if ((ret = OCIBindArrayOfStruct(bnd1p
			, hand.getError()
			, mxsize // pvskip
			, sizeof(sb2) // indskip
			, sizeof(ub2) // actual length skip parameter
			, 0)
		) != OCI_SUCCESS) {
			OracleExceptionFactory::throwit("OCIBindArrayOfStruct", hand, ret, _SQL);
		}
                ++pos;
        }
}

template<class TM, template<class> class SLMDC>
int SQLExecute<TM, SLMDC>::batchSize() const {
	if (batchParams.empty()) {
		return 1;
	} else {
		return batchParams.size();
	}
}

template<class TM, template<class> class SLMDC>
bool SQLExecute<TM, SLMDC>::push_onto_BatchParamSizes(typename super::super_bind::_argst& p) {
	if (batchParamData->sizes.size() != p.size()) {
		std::ostringstream os;
		os << "Error in batch execution: number of arguments mismatch. Current batch has " << batchParamData->sizes.size()  << " while the supplied statement has " << p.size();
		throw SQL::SQLException(os.str(), "07001", super::getSQL(), false);
		// note on usage of getSQL: when reaching this point it is already verified that this SQL equals that SQL
	}

	// we need to iterate twice..
	// first to check consistance of the actual types, without modifying any data in *this.
	//       in case they are not the same, we can still abort batching without 
	//       having corrupted data
	typename super::_argst::const_iterator end = p.end();
	typename super::_argst::const_iterator t =  super::_args.begin();
	for (typename super::super_bind::_argst::const_iterator i = p.begin(); i != end; ++i) {
		if ((*i)->type() != (*t)->type()) {
			// log an warning someting like:
			// std::cout << "Can not batch up similar statements due to different parameter type at position " << distance(static_cast<typename super::super_bind::_argst::const_iterator>(p.begin()), i) << " (" << ((*i)->type()) <<  " vs. " <<  ((*t)->type()) << ")" << std::endl;
			return false;
		}
		++t;
	}
	// the second iteration acutally takes over the parameters
        int pos = 0;
	for (typename super::super_bind::_argst::const_iterator i = p.begin(); i != end; ++i) {
		batchParamData->sizes[pos].push_back((*i)->size());
		batchParamData->indicators[pos].push_back((*i)->indicator());
		++pos;	
	}
	return true;
}

template<class TM, template<class> class SLMDC>
bool SQLExecute<TM, SLMDC>::addBatch(SQLExecute<TM, SLMDC>& p) {
	if (_sth.get()) {
		// TODO: recheck
		throw SQL::SQLException("building up a batch for a already prepared statement is invalid", "24000", p._SQL, false);
	}
	if (_SQL != p._SQL) {
		//std::cout << " false due to different sql (" << _SQL << ") vs. (" << p._SQL << ")" << std::endl;
		return false;
	}

	if (batchParams.empty()) {
		// if this is the first call to addBatch, we first copy the already
		// (normally) bound values into the first positon of our new structure
		batchParamData->sizes.resize(super::super_bind::_args.size());	
		batchParamData->indicators.resize(super::super_bind::_args.size());	

		batchParams.push_back(super::super_bind::_args);
		push_onto_BatchParamSizes(super::super_bind::_args);
	}
	if (push_onto_BatchParamSizes(p._args)) {
		batchParams.push_back(p._args);
		return true;
	} else {
		return false;
	}
}

template class SQLExecute<DefaultThreadedModel, oracle::SelectListMetaDataCache >;
template class SQLExecute<DefaultThreadedModel, oracle::SelectListMetaDataNopCache >;

} // namespace oracle
} // namespace fatalmind

